/* Some examples of integer casting
 * Bill Zorn <waz@andrew.cmu.edu>
 * 15-122 f12
 */

#include <stdio.h>
#include <stddef.h>
#include <inttypes.h>

int main()
{
  /* What happens when we cast from a 1-byte integer to a 4-byte integer?
   *
   * For clarity, we use the integer types from inttypes.h.
   * Note that int8_t is the same as a (signed) char, and int32_t is
   * (probably) the same as a (signed) int.
   *
   * Also for clarity, we use ridiculous formatting strings so that when
   * we print the values, they reflect the bit representations of the
   * corresponding integers. You do not need to understand how these
   * printf statements work.
   */

  int8_t c;
  int32_t i;

  uint8_t uc;
  uint32_t ui;

  int64_t l;

  /* We have an uninteresting case when c is positive. */
  c = 0x75;       // 117
  i = (int32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: 0x75 from int8_t to int32_t\n");
  printf("  c was: %#.2hhx, i is now: %#.8x\n\n", c, i);

  /* When c is negative, something strange happens to the high bits. */
  c = -0x70;      // -112, or 0x90, but writing it as 0x90 would be overflow
  i = (int32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: 0x90 from int8_t to int32_t\n");
  printf("  c was: %#.2hhx, i is now: %#.8x\n\n", c, i);

  /* Since there are more bits in an int32_t than there are in an int8_t,
   * something must happen to the high bits. If we make a signed cast,
   * then they are all set to the same value as the sign bit of the integer
   * we are casting from. Note that this preserves the signed value of the
   * number.
   */

  /* As above, with decimal values. */
  c = 117;        // 0x75
  i = (int32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: 117 from int8_t to int32_t\n");
  printf("  c was: %hhd, i is now: %d\n\n", c, i);

  /* As above, with decimal values. */
  c = -112;       // 0x90
  i = (int32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: -112 from int8_t to int32_t\n");
  printf("  c was: %hhd, i is now: %d\n\n", c, i);

  /* If our integer types are not signed, then the high bits are always
   * filled with 0, as this correctly preserves the numerical value.
   */

  /* We now perform the same casts using unsigned integers. */
  uc = 0x75;         // 117
  ui = (uint32_t)uc; // cast 1-byte to 4-byte, unsigned

  printf("casting: 0x75 from uint8_t to uint32_t\n");
  printf("  uc was: %#.2hhx, ui is now: %#.8x\n\n", uc, ui);

  /* The strange sign extension behavior is not observed. */
  uc = 0x90;         // 144
  ui = (uint32_t)uc; // cast 1-byte to 4-byte, unsigned

  printf("casting: 0x90 from uint8_t to uint32_t\n");
  printf("  uc was: %#.2hhx, ui is now: %#.8x\n\n", uc, ui);

  /* Finally, note that the sign extension behavior is only determined by
   * the type we are casting from. If we are casting from a signed type, then
   * we will always sign extend with the value of the sign bit; if we are
   * casting from unsigned, we will always sign extend with 0. The destination
   * type (specifically, whether or not is is signed) does not matter.
   */

  /* Signed to unsigned: still uninteresting. */
  c = 0x75;         // 117
  ui = (uint32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: 0x75 from int8_t to uint32_t\n");
  printf("  c was: %#.2hhx, ui is now: %#.8x\n\n", c, ui);

  /* The same as the signed to signed behavior. */
  c = -0x70;        // -112, or 0x90
  ui = (uint32_t)c; // cast 1-byte to 4-byte, signed

  printf("casting: 0x90 from int8_t to uint32_t\n");
  printf("  c was: %#.2hhx, ui is now: %#.8x\n\n", c, ui);

  /* Unsigned to signed. */
  uc = 0x75;       // 117
  i = (int32_t)uc; // cast 1-byte to 4-byte, unsigned

  printf("casting: 0x75 from uint8_t to int32_t\n");
  printf("  uc was: %#.2hhx, i is now: %#.8x\n\n", uc, i);

  /* Same as unsigned to unsigned. */
  uc = 0x90;       // 144
  i = (int32_t)uc; // cast 1-byte to 4-byte, unsigned

  printf("casting: 0x90 from uint8_t to int32_t\n");
  printf("  uc was: %#.2hhx, i is now: %#.8x\n\n", uc, i);

  /* As a last aside, we note that casting within a single integer size
   * has no impact on the binary representation of the integer. The only
   * difference is how the integer will be treated by other operations,
   * such as other casts or comparisons.
   *
   * Also, we should note that casting to a smaller integer type will
   * always truncate the integer by simply chopping off the bits on the
   * left. We don't really need examples, as the behavior is consistent,
   * but be aware that it might destroy information about the number.
   * Casting to a larger type does not have this property: we can always
   * cast back down to get the same number. However, if we cast down and
   * then back up, the high bits will be set either all to 0 or all to 1.
   */

  /* deadbeef is deadbeef, signed or unsigned. */
  ui = 0xdeadbeef;
  i = (int32_t)ui;

  printf("casting: 0xdeadbeef from uint32_t to int32_t\n");
  printf("  ui was: %#.8x, i is now: %#.8x\n\n", ui, i);

  /* Cast from unsigned. */
  l = (int64_t)ui;

  printf("casting: 0xdeadbeef from uint32_t to int64_t\n");
  printf("  ui was: %#.8x, l is now: %#.16lx\n\n", ui, l);

  /* Different when we make the same cast from signed. */
  l = (int64_t)i;

  printf("casting: 0xdeadbeef from int32_t to int64_t\n");
  printf("  i was: %#.8x, l is now: %#.16lx\n\n", i, l);

  return 0;
}
